@@ -15,8 +15,8 @@ except ImportError:
 async_files = ['test_async.py', 'asyncio.py']
 
 
-def pytest_ignore_collect(path):
-    """ Text collection function executed by pytest"""
-    if not WITH_ASYNC and basename(str(path)) in async_files:
+def pytest_ignore_collect(collection_path):
+    """Text collection function executed by pytest"""
+    if not WITH_ASYNC and basename(str(collection_path)) in async_files:
         return True
     return False
