@@ -12,7 +12,7 @@ from unittest import skipIf
 from functools import partial
 import weakref
 from .test_core import TestTransitions, MachineError, TYPE_CHECKING
-from .utils import DummyModel
+from .utils import DummyModel, Stuff
 from .test_graphviz import pgv as gv
 from .test_pygraphviz import pgv
 
@@ -435,6 +435,21 @@ class TestAsync(TestTransitions):
 
         asyncio.run(run())
 
+    def test_on_exception_finalize(self):
+        mock = MagicMock()
+
+        def finalize():
+            mock()
+            raise RuntimeError("Could not finalize")
+
+        m = self.machine_cls(states=['A', 'B'], initial='A', finalize_event=finalize)
+
+        async def run():
+            self.assertTrue(await m.to_B())
+            self.assertTrue(mock.called)
+
+        asyncio.run(run())
+
     def test_weakproxy_model(self):
         d = DummyModel()
         pr = weakref.proxy(d)
@@ -483,6 +498,65 @@ class TestAsync(TestTransitions):
 
         asyncio.run(run())
 
+    def test_may_transition_internal(self):
+        states = ['A', 'B', 'C']
+        d = DummyModel()
+        _ = self.machine_cls(model=d, states=states, transitions=[["go", "A", "B"], ["wait", "B", None]],
+                             initial='A', auto_transitions=False)
+
+        async def run():
+            assert await d.may_go()
+            assert not await d.may_wait()
+            await d.go()
+            assert not await d.may_go()
+            assert await d.may_wait()
+
+        asyncio.run(run())
+
+    def test_may_transition_with_exception(self):
+
+        stuff = Stuff(machine_cls=self.machine_cls, extra_kwargs={"send_event": True})
+        stuff.machine.add_transition(trigger="raises", source="A", dest="B", prepare=partial(stuff.this_raises, RuntimeError("Prepare Exception")))
+        stuff.machine.add_transition(trigger="raises", source="B", dest="C", conditions=partial(stuff.this_raises, ValueError("Condition Exception")))
+        stuff.machine.add_transition(trigger="works", source="A", dest="B")
+
+        def process_exception(event_data):
+            assert event_data.error is not None
+            assert event_data.transition is not None
+            assert event_data.event.name == "raises"
+            assert event_data.machine == stuff.machine
+
+        async def run():
+            with self.assertRaises(RuntimeError):
+                await stuff.may_raises()
+            assert stuff.is_A()
+            assert await stuff.may_works()
+            assert await stuff.works()
+            with self.assertRaises(ValueError):
+                await stuff.may_raises()
+            assert stuff.is_B()
+            stuff.machine.on_exception.append(process_exception)
+            assert not await stuff.may_raises()
+            assert await stuff.to_A()
+            assert not await stuff.may_raises()
+
+        asyncio.run(run())
+
+    def test_on_final(self):
+        final_mock = MagicMock()
+        machine = self.machine_cls(states=['A', {'name': 'B', 'final': True}], on_final=final_mock, initial='A')
+
+        async def run():
+            self.assertFalse(final_mock.called)
+            await machine.to_B()
+            self.assertTrue(final_mock.called)
+            await machine.to_A()
+            self.assertEqual(1, final_mock.call_count)
+            await machine.to_B()
+            self.assertEqual(2, final_mock.call_count)
+
+        asyncio.run(run())
+
 
 @skipIf(asyncio is None or (pgv is None and gv is None), "AsyncGraphMachine requires asyncio and (py)gaphviz")
 class TestAsyncGraphMachine(TestAsync):
@@ -530,6 +604,49 @@ class TestHierarchicalAsync(TestAsync):
         asyncio.run(machine.to_B())
         self.assertTrue(machine.is_B())
 
+    def test_final_state_nested(self):
+        final_mock_B = MagicMock()
+        final_mock_Y = MagicMock()
+        final_mock_Z = MagicMock()
+        final_mock_machine = MagicMock()
+
+        mocks = [final_mock_B, final_mock_Y, final_mock_Z, final_mock_machine]
+
+        states = ['A', {'name': 'B', 'parallel': [{'name': 'X', 'final': True},
+                                                  {'name': 'Y', 'transitions': [['final_Y', 'yI', 'yII']],
+                                                   'initial': 'yI',
+                                                   'on_final': final_mock_Y,
+                                                   'states':
+                                                       ['yI', {'name': 'yII', 'final': True}]
+                                                   },
+                                                  {'name': 'Z', 'transitions': [['final_Z', 'zI', 'zII']],
+                                                   'initial': 'zI',
+                                                   'on_final': final_mock_Z,
+                                                   'states':
+                                                       ['zI', {'name': 'zII', 'final': True}]
+                                                   },
+                                                  ],
+                        "on_final": final_mock_B}]
+
+        machine = self.machine_cls(states=states, on_final=final_mock_machine, initial='A')
+
+        async def run():
+            self.assertFalse(any(mock.called for mock in mocks))
+            await machine.to_B()
+            self.assertFalse(any(mock.called for mock in mocks))
+            await machine.final_Y()
+            self.assertTrue(final_mock_Y.called)
+            self.assertFalse(final_mock_Z.called)
+            self.assertFalse(final_mock_B.called)
+            self.assertFalse(final_mock_machine.called)
+            await machine.final_Z()
+            self.assertEqual(1, final_mock_Y.call_count)
+            self.assertEqual(1, final_mock_Z.call_count)
+            self.assertEqual(1, final_mock_B.call_count)
+            self.assertEqual(1, final_mock_machine.call_count)
+
+        asyncio.run(run())
+
 
 @skipIf(asyncio is None or (pgv is None and gv is None), "AsyncGraphMachine requires asyncio and (py)gaphviz")
 class TestAsyncHierarchicalGraphMachine(TestHierarchicalAsync):
