@@ -930,13 +930,15 @@ class TestTransitions(TestCase):
         def check_prepare_repr(event_data):
             self.assertRegex(
                 str(event_data),
-                r"<EventData\('<State\('A'\)@\d+>', "
+                r"<EventData\(<Event\('do_strcheck'\)@\d+>, "
+                r"<State\('A'\)@\d+>, "
                 r"None\)@\d+>")
 
         def check_before_repr(event_data):
             self.assertRegex(
                 str(event_data),
-                r"<EventData\('<State\('A'\)@\d+>', "
+                r"<EventData\(<Event\('do_strcheck'\)@\d+>, "
+                r"<State\('A'\)@\d+>, "
                 r"<Transition\('A', 'B'\)@\d+>\)@\d+>")
             m.checked = True
 
@@ -1229,12 +1231,15 @@ class TestTransitions(TestCase):
         m = Machine(model=d, states=states, initial='A', auto_transitions=False)
         m.add_transition('walk', 'A', 'B')
         m.add_transition('stop', 'B', 'C')
+        m.add_transition('wait', 'B', None)
         assert d.may_walk()
         assert not d.may_stop()
+        assert not d.may_wait()
 
         d.walk()
         assert not d.may_walk()
         assert d.may_stop()
+        assert d.may_wait()
 
     def test_may_transition_for_autogenerated_triggers(self):
         states = ['A', 'B', 'C']
@@ -1289,3 +1294,39 @@ class TestTransitions(TestCase):
         m = self.machine_cls(model=d, states=states, initial='A', auto_transitions=False)
         m.add_transition('walk', 'A', 'UNKNOWN')
         assert not d.may_walk()
+
+    def test_may_transition_with_exception(self):
+        stuff = Stuff(machine_cls=self.machine_cls, extra_kwargs={"send_event": True})
+        stuff.machine.add_transition(trigger="raises", source="A", dest="B", prepare=partial(stuff.this_raises, RuntimeError("Prepare Exception")))
+        stuff.machine.add_transition(trigger="raises", source="B", dest="C", conditions=partial(stuff.this_raises, ValueError("Condition Exception")))
+        stuff.machine.add_transition(trigger="works", source="A", dest="B")
+
+        def process_exception(event_data):
+            assert event_data.error is not None
+            assert event_data.transition is not None
+            assert event_data.event.name == "raises"
+            assert event_data.machine == stuff.machine
+
+        with self.assertRaises(RuntimeError):
+            stuff.may_raises()
+        assert stuff.is_A()
+        assert stuff.may_works()
+        assert stuff.works()
+        with self.assertRaises(ValueError):
+            stuff.may_raises()
+        assert stuff.is_B()
+        stuff.machine.on_exception.append(process_exception)
+        assert not stuff.may_raises()
+        assert stuff.to_A()
+        assert not stuff.may_raises()
+
+    def test_on_final(self):
+        final_mock = MagicMock()
+        machine = self.machine_cls(states=['A', {'name': 'B', 'final': True}], on_final=final_mock, initial='A')
+        self.assertFalse(final_mock.called)
+        machine.to_B()
+        self.assertTrue(final_mock.called)
+        machine.to_A()
+        self.assertEqual(1, final_mock.call_count)
+        machine.to_B()
+        self.assertEqual(2, final_mock.call_count)
