@@ -1,20 +1,16 @@
-try:
-    from builtins import object
-except ImportError:
-    pass
-
 from .utils import Stuff, DummyModel
 from .test_core import TestTransitions, TYPE_CHECKING
 
 from transitions.extensions import (
     LockedGraphMachine, GraphMachine, HierarchicalGraphMachine, LockedHierarchicalGraphMachine
 )
-from transitions.extensions.nesting import NestedState
 from transitions.extensions.states import add_state_features, Timeout, Tags
 from unittest import skipIf
 import tempfile
 import os
 import re
+import sys
+from unittest import TestCase
 
 try:
     # Just to skip tests if graphviz not installed
@@ -22,10 +18,23 @@ try:
 except ImportError:  # pragma: no cover
     pgv = None
 
-
 if TYPE_CHECKING:
     from typing import Type, List, Collection, Union
 
+edge_re = re.compile(r"^\s+(?P<src>\w+)\s*->\s*(?P<dst>\w+)\s*(?P<attr>\[.*\]?)\s*$")
+node_re = re.compile(r"^\s+(?P<node>\w+)\s+(?P<attr>\[.*\]?)\s*$")
+
+
+class TestDiagramsImport(TestCase):
+
+    use_pygraphviz = False
+    pgv = pgv
+
+    def test_import(self):
+        machine = GraphMachine(None, use_pygraphviz=self.use_pygraphviz)
+        if machine.graph_cls is None:
+            self.assertIsNone(pgv)
+
 
 @skipIf(pgv is None, 'Graph diagram test requires graphviz.')
 class TestDiagrams(TestTransitions):
@@ -38,16 +47,19 @@ class TestDiagrams(TestTransitions):
             dot = graph.string()
         else:
             dot = graph.source
-        nodes = []
+        nodes = set()
         edges = []
         for line in dot.split('\n'):
-            if '->' in line:
-                src, rest = line.split('->')
-                dst, attr = rest.split(None, 1)
-                nodes.append(src.strip().replace('"', ''))
-                nodes.append(dst)
-                edges.append(attr)
-        return dot, set(nodes), edges
+            match = edge_re.search(line)
+            if match:
+                nodes.add(match.group("src"))
+                nodes.add(match.group("dst"))
+                edges.append(match.group("attr"))
+            else:
+                match = node_re.search(line)
+                if match and match.group("node") not in ["node", "graph", "edge"]:
+                    nodes.add(match.group("node"))
+        return dot, nodes, edges
 
     def tearDown(self):
         pass
@@ -206,8 +218,8 @@ class TestDiagrams(TestTransitions):
         m.to_B()
         g3 = m.get_graph(show_roi=True)
         _, nodes, edges = self.parse_dot(g3)
-        self.assertEqual(len(edges), 3)
-        self.assertEqual(len(nodes), 4)
+        self.assertEqual(len(edges), 3)  # to_state_{A,C,F}
+        self.assertEqual(len(nodes), 5)  # B + A,C,F (edges) + E (previous)
 
     def test_state_tags(self):
 
@@ -266,17 +278,37 @@ class TestDiagrams(TestTransitions):
                 m = self.machine_cls(states=['A', 'B', 'C'], initial='A', use_pygraphviz=True)
             # make sure to reload after test is done to avoid side effects with other tests
             reload(diagrams_pygraphviz)
-            print(m.graph_cls, pgv)
+            # print(m.graph_cls, pgv)
             self.assertTrue(issubclass(m.graph_cls, Graph))
         except ImportError:
             pass
 
+    def test_function_callbacks_annotation(self):
+        m = self.machine_cls(states=['A', 'B'], initial='A', use_pygraphviz=self.use_pygraphviz, show_conditions=True)
+        m.add_transition('advance', 'A', 'B', conditions=m.is_A, unless=m.is_B)
+        _, nodes, edges = self.parse_dot(m.get_graph())
+        self.assertIn("[is_state(A", edges[0])
+
+    def test_update_on_remove_transition(self):
+        m = self.machine_cls(states=self.states, transitions=self.transitions, initial='A',
+                             use_pygraphviz=self.use_pygraphviz, show_state_attributes=True)
+        _, _, edges = self.parse_dot(m.get_graph())
+        assert "[label=walk]" in edges
+        m.remove_transition(trigger="walk", source="A", dest="B")
+        _, _, edges = self.parse_dot(m.get_graph())
+        assert not any("walk" == t["trigger"] for t in m.markup["transitions"])
+        assert "[label=walk]" not in edges
+
 
 @skipIf(pgv is None, 'Graph diagram test requires graphviz')
 class TestDiagramsLocked(TestDiagrams):
 
     machine_cls = LockedGraphMachine  # type: Type[LockedGraphMachine]
 
+    @skipIf(sys.version_info < (3, ), "Python 2.7 cannot retrieve __name__ from partials")
+    def test_function_callbacks_annotation(self):
+        super(TestDiagramsLocked, self).test_function_callbacks_annotation()
+
 
 @skipIf(pgv is None, 'NestedGraph diagram test requires graphviz')
 class TestDiagramsNested(TestDiagrams):
@@ -308,8 +340,7 @@ class TestDiagramsNested(TestDiagrams):
 
         self.assertEqual(len(edges), 8)
         # Test that graph properties match the Machine
-        self.assertEqual(set(m.states.keys()) - set(['C', 'C%s1' % NestedState.separator]),
-                         set(nodes) - set(['C_anchor', 'C%s1_anchor' % NestedState.separator]))
+        self.assertEqual(set(m.get_nested_state_names()), nodes)
         m.walk()
         m.run()
 
@@ -344,6 +375,50 @@ class TestDiagramsNested(TestDiagrams):
         self.assertEqual(len(edges), 2)
         self.assertEqual(len(nodes), 3)
 
+    def test_roi_parallel(self):
+        class Model:
+            @staticmethod
+            def is_fast(*args, **kwargs):
+                return True
+
+        self.states[0] = {"name": "A", "parallel": ["1", "2"]}
+
+        model = Model()
+        m = self.machine_cls(model, states=self.states, transitions=self.transitions, initial='A', title='A test',
+                             use_pygraphviz=self.use_pygraphviz, show_conditions=True)
+        g1 = model.get_graph(show_roi=True)
+        _, nodes, edges = self.parse_dot(g1)
+        self.assertEqual(len(edges), 2)  # reset and walk
+        print(nodes)
+        self.assertEqual(len(nodes), 4)
+        model.walk()
+        model.run()
+        model.sprint()
+        g2 = model.get_graph(show_roi=True)
+        dot, nodes, edges = self.parse_dot(g2)
+        self.assertEqual(len(edges), 2)
+        self.assertEqual(len(nodes), 3)
+
+    def test_roi_parallel_deeper(self):
+        states = ['A', 'B', 'C', 'D',
+                  {'name': 'P',
+                   'parallel': [
+                       '1',
+                       {'name': '2', 'parallel': [
+                           {'name': 'a'},
+                           {'name': 'b', 'parallel': [
+                               {'name': 'x', 'parallel': ['1', '2']}, 'y'
+                           ]}
+                       ]},
+                   ]}]
+        transitions = [["go", "A", "P"], ["reset", "*", "A"]]
+        m = self.machine_cls(states=states, transitions=transitions, initial='A', title='A test',
+                             use_pygraphviz=self.use_pygraphviz, show_conditions=True)
+        m.go()
+        _, nodes, edges = self.parse_dot(m.get_graph(show_roi=True))
+        self.assertEqual(len(edges), 2)
+        self.assertEqual(len(nodes), 10)
+
     def test_internal(self):
         states = ['A', 'B']
         transitions = [['go', 'A', 'B'],
@@ -353,6 +428,7 @@ class TestDiagramsNested(TestDiagrams):
                              use_pygraphviz=self.use_pygraphviz)
 
         _, nodes, edges = self.parse_dot(m.get_graph())
+        print(nodes)
         self.assertEqual(len(nodes), 2)
         self.assertEqual(len([e for e in edges if '[internal]' in e]), 1)
 
@@ -435,3 +511,7 @@ class TestDiagramsLockedNested(TestDiagramsNested):
     def setUp(self):
         super(TestDiagramsLockedNested, self).setUp()
         self.machine_cls = LockedHierarchicalGraphMachine  # type: Type[LockedHierarchicalGraphMachine]
+
+    @skipIf(sys.version_info < (3, ), "Python 2.7 cannot retrieve __name__ from partials")
+    def test_function_callbacks_annotation(self):
+        super(TestDiagramsLockedNested, self).test_function_callbacks_annotation()
