@@ -720,6 +720,45 @@ class TestNestedTransitions(TestTransitions):
         machine.reflexive()
         self.assertEqual(state_name, machine.state)
 
+    def test_final_state_nested(self):
+        final_mock_B = MagicMock()
+        final_mock_Y = MagicMock()
+        final_mock_Z = MagicMock()
+        final_mock_machine = MagicMock()
+
+        mocks = [final_mock_B, final_mock_Y, final_mock_Z, final_mock_machine]
+
+        states = ['A', {'name': 'B', 'parallel': [{'name': 'X', 'final': True},
+                                                  {'name': 'Y', 'transitions': [['final_Y', 'yI', 'yII']],
+                                                   'initial': 'yI',
+                                                   'on_final': final_mock_Y,
+                                                   'states':
+                                                       ['yI', {'name': 'yII', 'final': True}]
+                                                   },
+                                                  {'name': 'Z', 'transitions': [['final_Z', 'zI', 'zII']],
+                                                   'initial': 'zI',
+                                                   'on_final': final_mock_Z,
+                                                   'states':
+                                                       ['zI', {'name': 'zII', 'final': True}]
+                                                   },
+                                                  ],
+                        "on_final": final_mock_B}]
+
+        machine = self.machine_cls(states=states, on_final=final_mock_machine, initial='A')
+        self.assertFalse(any(mock.called for mock in mocks))
+        machine.to_B()
+        self.assertFalse(any(mock.called for mock in mocks))
+        machine.final_Y()
+        self.assertTrue(final_mock_Y.called)
+        self.assertFalse(final_mock_Z.called)
+        self.assertFalse(final_mock_B.called)
+        self.assertFalse(final_mock_machine.called)
+        machine.final_Z()
+        self.assertEqual(1, final_mock_Y.call_count)
+        self.assertEqual(1, final_mock_Z.call_count)
+        self.assertEqual(1, final_mock_B.call_count)
+        self.assertEqual(1, final_mock_machine.call_count)
+
 
 class TestSeparatorsBase(TestCase):
 
@@ -738,6 +777,17 @@ class TestSeparatorsBase(TestCase):
         self.state_cls = CustomNestedState
         self.stuff = Stuff(self.states, self.machine_cls)
 
+    def test_add_nested_state(self):
+        m = self.machine_cls(states=['A'], initial='A')
+        m.add_state('B{0}1{0}a'.format(self.state_cls.separator))
+        m.add_state('B{0}2{0}b'.format(self.state_cls.separator))
+        self.assertIn('B', m.states)
+        self.assertIn('1', m.states['B'].states)
+        self.assertIn('a', m.states['B'].states['1'].states)
+
+        with self.assertRaises(ValueError):
+            m.add_state(m.states['A'])
+
     def test_enter_exit_nested(self):
         separator = self.state_cls.separator
         s = self.stuff
@@ -899,6 +949,7 @@ class TestSeparatorsBase(TestCase):
         transitions = [
             {'trigger': 'walk', 'source': 'A', 'dest': 'B'},
             {'trigger': 'run', 'source': 'B', 'dest': 'C'},
+            {'trigger': 'wait', 'source': 'B', 'dest': None},
             {'trigger': 'run_fast', 'source': 'C', 'dest': 'C{0}1'.format(self.separator)},
             {'trigger': 'sprint', 'source': 'C', 'dest': 'D'}
         ]
@@ -909,17 +960,38 @@ class TestSeparatorsBase(TestCase):
         assert not m.may_run()
         assert not m.may_run_fast()
         assert not m.may_sprint()
+        assert not m.may_wait()
 
         m.walk()
         assert not m.may_walk()
         assert m.may_run()
         assert not m.may_run_fast()
+        assert m.may_wait()
 
         m.run()
         assert m.may_run_fast()
         assert m.may_sprint()
         m.run_fast()
 
+    def test_remove_nested_transition(self):
+        separator = self.state_cls.separator
+        state = {
+            'name': 'B',
+            'children': ['1', '2'],
+            'transitions': [['jo', '1', '2']],
+            'initial': '2'
+        }
+        m = self.stuff.machine_cls(initial='A', states=['A', state],
+                                   transitions=[['go', 'A', 'B'], ['go', 'B{0}2'.format(separator),
+                                                                   'B{0}1'.format(separator)]])
+        m.remove_transition("go", "A", "B")
+        assert m.get_transitions("go")
+        m.remove_transition("go")
+        assert not m.get_transitions("go")
+        assert m.get_transitions("jo")
+        m.remove_transition("jo")
+        assert not m.get_transitions("jo")
+
 
 class TestSeparatorsSlash(TestSeparatorsBase):
     separator = '/'
